/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/opengl/opengl.h: Interface to OpenGL for use by system-specific
 * graphics implementations.
 */

#ifndef SIL_SRC_SYSDEP_OPENGL_OPENGL_H
#define SIL_SRC_SYSDEP_OPENGL_OPENGL_H

#include "src/sysdep/opengl/gl-headers.h"

/*************************************************************************/
/*************** OpenGL feature flags (for opengl_init()) ****************/
/*************************************************************************/

/* GL object deletion should be delayed until opengl_free_dead_resources()
 * is called.  (Set this flag if delete operations tend to block for long
 * periods of time.) */
#define OPENGL_FEATURE_DELAYED_DELETE      (1<<0)

/* Vertex buffer objects for dynamic (immediate-mode) primitives are
 * faster than client-side data buffers, or client-side data buffers are
 * not available.  This flag is ignored if shaders are not available. */
#define OPENGL_FEATURE_FAST_DYNAMIC_VBO    (1<<1)

/* Vertex buffer objects for static (create-once, draw-many) primitives
 * are faster than client-side data buffers, or client-side data buffers
 * are not available.  This flag is ignored if shaders are not available. */
#define OPENGL_FEATURE_FAST_STATIC_VBO     (1<<2)

/* The GL has framebuffer support (OpenGL 3.0 or EXT_framebuffer_object,
 * OpenGL ES 1.1). */
#define OPENGL_FEATURE_FRAMEBUFFERS        (1<<3)

/* The glGenerateMipmap() function should be used for generating mipmaps,
 * instead of manually generating mipmaps from the pixel data.  Since
 * glGenerateMipmap() is part of the framebuffer extension to OpenGL, this
 * flag only makes sense when used in combination with
 * OPENGL_FEATURE_FRAMEBUFFERS. */
#define OPENGL_FEATURE_GENERATEMIPMAP      (1<<4)

/* Vertex array objects are required for primitive rendering.  This should
 * be set when using a core profile implementation of OpenGL 3.0 or later. */
#define OPENGL_FEATURE_MANDATORY_VAO       (1<<5)

/* The GL and graphics hardware can draw quads and quad strips natively.
 * (If not, quads will be drawn as indexed triangles.)  Note that this
 * flag should not be used with an OpenGL 3.0 (or later) core profile
 * renderer, since OpenGL 3.0 deprecates quads entirely. */
#define OPENGL_FEATURE_NATIVE_QUADS        (1<<6)

/* GL has shader support (OpenGL 2.0, OpenGL ES 2.0).  This assumes support
 * for vertex buffer objects as well, though VBOs are not used unless one
 * or both of the FAST_*_VBO feature flags are also set. */
#define OPENGL_FEATURE_SHADERS             (1<<7)

/* Vertex array objects should be used for static, but not dynamic,
 * primitive rendering.  This flag has no meaning if the MANDATORY_VAO
 * feature flag is set. */
#define OPENGL_FEATURE_USE_STATIC_VAO      (1<<8)


/* The following flags are set automatically based on the OpenGL version
 * reported by the renderer. */


/* The glCompressedTexImage() function is available (OpenGL 1.3 or
 * ARB_texture_compression, OpenGL ES 1.1). */
#define OPENGL_FEATURE_COMPRESSEDTEXIMAGE  (1<<16)

/* The glDiscardFramebufferEXT() function is available (OpenGL ES 2.x with
 * EXT_discard_framebuffer). */
#define OPENGL_FEATURE_DISCARD_FRAMEBUFFER (1<<17)

/* The fixed-function pipeline for primitive rendering is available
 * (OpenGL 1.x/2.x or compatibility profile, OpenGL ES 1.x). */
#define OPENGL_FEATURE_FFP_PRIMITIVES      (1<<18)

/* The glGetTexImage() function is available. */
#define OPENGL_FEATURE_GETTEXIMAGE         (1<<19)

/* All automatically-set flags. */
#define OPENGL_AUTOCONFIG_FEATURE_MASK     (0xFFFF0000U)


/*************************************************************************/
/***** OpenGL texture/index format flags (for opengl_has_formats()) ******/
/*************************************************************************/

/* BGRA ordering is supported for RGB pixel data. */
#define OPENGL_FORMAT_BGRA      (1<<0)

/* Reversed bit ordering (GL_..._REV) is supported for packed pixel formats. */
#define OPENGL_FORMAT_BITREV    (1<<1)

/* Alpha-only textures are supported. */
#define OPENGL_FORMAT_ALPHA     (1<<2)

/* S3TC compression formats (DXTn) are supported. */
#define OPENGL_FORMAT_S3TC      (1<<3)

/* PVRTC compression formats are supported. */
#define OPENGL_FORMAT_PVRTC     (1<<4)

/* 32-bit index values are supported for indexed primitives. */
#define OPENGL_FORMAT_INDEX32   (1<<5)

/*************************************************************************/
/************************ Callback function types ************************/
/*************************************************************************/

/**
 * ShaderSourceCallback:  Function type for the shader source generation
 * function passed to opengl_set_shader_callbacks().  The function should
 * generate and load (using glShaderSource()) source code for the vertex
 * and fragment shaders specified by the parameters to the function, which
 * indicate the types of data included in each vertex and other relevant
 * render state.
 *
 * Vertex shaders must use the following names for vertex attributes and
 * uniform values:
 *
 *       transform: (uniform mat4) Model-view-projection transformation matrix
 *         texture: (uniform sampler) Texture sampler (may be any sampler type)
 *      tex_offset: (uniform vec2) Texture offset for all vertices
 *     fixed_color: (uniform vec4) Fixed color multiplier for all vertices
 *       fog_depth: (uniform vec2) Fog start and end positions (in x and y)
 *       fog_color: (uniform vec4) Fog color
 *       alpha_ref: (uniform float) Reference value for alpha testing
 *        position: (attribute vec[234]) Vertex position
 *        texcoord: (attribute vec2) Texture coordinates for vertex
 *           color: (attribute vec4) Color for vertex
 *
 * [Parameters]
 *     vertex_shader: OpenGL vertex shader object.
 *     fragment_shader: OpenGL fragment shader object.
 *     primitive_type: Primitive type (GL_TRIANGLES, etc.).
 *     position_count: Number of position elements (2-4).
 *     texcoord_count: Number of texture coordinate elements (0 or 2).
 *     texcolor_type: Texture color type (GL_RGBA or GL_ALPHA, or 0 for
 *         OpenGL  external textures); ignored if texturing is disabled.
 *     tex_offset: True for an external texture offset passed as a uniform
 *         parameter, false for none.
 *     color_count: Number of color elements (0 or 4).
 *     color_uniform: True for an external fixed color passed as a uniform
 *         parameter, false for none.
 *     fog: True for linear fog, false for no fog.
 *     alpha_test: True for alpha testing (discard pixels with alpha less
 *         than the reference value), false for no alpha testing.
 *     alpha_comparison: Alpha test comparison type (GL_GEQUAL, etc.).
 * [Return value]
 *     True on success, false on error.
 */
typedef int ShaderSourceCallback(
    GLuint vertex_shader, GLuint fragment_shader,
    GLenum primitive_type, int position_count, int texcoord_count,
    GLenum texcolor_type, int tex_offset, int color_count,
    int color_uniform, int fog, int alpha_test, GLenum alpha_comparison);

/**
 * ShaderKeyCallback:  Function type for the shader key generation function
 * passed to opengl_set_shader_callbacks().  The function must return a
 * 32-bit value, unique among all possible shaders, identifying the shader
 * specified by the function parameters (which are the same values as would
 * be passed to the shader source callback to generate the shader's source).
 *
 * If the shader source generator cannot handle the requested parameters,
 * it should return INVALID_SHADER_KEY; in this case, primitives using
 * those parameters will be unable to be rendered.
 *
 * [Parameters]
 *     As for ShaderSourceCallback, but does not take vertex_shader or
 *     fragment_shader.
 * [Return value]
 *     Unique 32-bit shader key, or INVALID_SHADER_KEY if the requested
 *     parameters are not supported.
 */
typedef uint32_t ShaderKeyCallback(
    GLenum primitive_type, int position_count, int texcoord_count,
    GLenum texcolor_type, int tex_offset, int color_count,
    int color_uniform, int fog, int alpha_test, GLenum alpha_comparison);

#define INVALID_SHADER_KEY (~(uint32_t)0)

/*************************************************************************/
/************************* Convenience functions *************************/
/*************************************************************************/

/**
 * opengl_clear_error:  Clear any pending GL error.  Equivalent to looping
 * over glGetError() until it returns GL_NO_ERROR, but makes the intent
 * clearer.
 */
static inline void opengl_clear_error(void)
{
    while (glGetError() != GL_NO_ERROR) { /*spin*/ }
}

/*************************************************************************/
/******************** Interface function declarations ********************/
/*************************************************************************/

/******** framebuffer.c ********/

/**
 * opengl_set_default_framebuffer:  Set the default framebuffer ID to be
 * used when rendering to the display.  If not set, this defaults to zero.
 * (This value is meaningless if framebuffers are not available.)
 *
 * [Parameters]
 *     default_fb: Default framebuffer ID.
 */
extern void opengl_set_default_framebuffer(GLuint default_fb);

/**
 * opengl_get_default_framebuffer:  Return the default framebuffer ID as
 * set by opengl_set_default_framebuffer().
 *
 * [Return value]
 *     Default framebuffer ID.
 */
extern GLuint opengl_get_default_framebuffer(void);

/**
 * opengl_set_depth_bits, opengl_set_stencil_bits:  Set the desired bit
 * depth for depth and stencil buffers attached to framebuffers.
 * (System-specific code is responsible for setting the proper depth
 * and stencil buffer parameters on the display framebuffer.)  Currently,
 * the following combinations are supported:
 *
 * depth <= 16, stencil == 0 -- 16-bit depth buffer
 * depth <= 24, stencil == 0 -- 24-bit depth buffer
 * depth <= 24, stencil <= 8 -- 24-bit depth buffer and 8-bit stencil buffer
 * depth <= 32, stencil == 0 -- 32-bit depth buffer
 * depth <= 32, stencil <= 8 -- 32-bit floating-point depth buffer and 8-bit
 *                                 stencil buffer (not supported in OpenGL ES)
 *
 * Requesting an unsupported bit depth will cause all framebuffer creation
 * operations to fail.
 *
 * These values can be set before calling opengl_init(), and they are not
 * reset by any functions in the OpenGL module.
 *
 * [Parameters]
 *     bits: Minimum number of depth or stencil bits.
 */
extern void opengl_set_depth_bits(int bits);
extern void opengl_set_stencil_bits(int bits);


/******** graphics.c ********/

/**
 * opengl_init:  Initialize the OpenGL interface.  This does _not_ set up
 * the display or GL context; those tasks are the responsibility of
 * system-dependent code, and must be completed before this function is
 * called.
 *
 * [Parameters]
 *     width, height: Size of the OpenGL output window, in pixels.
 *     features: Bitmask of available OpenGL features (OPENGL_FEATURE_*).
 * [Return value]
 *     True on success, false on error.
 */
extern int opengl_init(int width, int height, int features);

/**
 * opengl_cleanup:  Clean up and shut down the OpenGL interface.
 */
extern void opengl_cleanup(void);

/**
 * opengl_major_version:  Return the major version number of OpenGL
 * supported by the runtime environment.  For example:
 *    - OpenGL 4.3 => return value 4
 *    - OpenGL ES 2.0 => return value 2
 *
 * [Return value]
 *     OpenGL major version number.
 */
extern int opengl_major_version(void);

/**
 * opengl_minor_version:  Return the minor version number of OpenGL
 * supported by the runtime environment.  For example:
 *    - OpenGL 4.3 => return value 3
 *    - OpenGL ES 2.0 => return value 0
 *
 * [Return value]
 *     OpenGL minor version number.
 */
extern int opengl_minor_version(void);

/**
 * opengl_has_extension:  Return whether the OpenGL implementation supports
 * the given extension.  Used internally by other OpenGL functions to check
 * feature availability, but may also be called from outside.
 *
 * [Parameters]
 *     name: Extension name, including the "GL_" prefix.
 * [Return value]
 *     True if the extension is supported, false if not.
 */
extern PURE_FUNCTION int opengl_has_extension(const char *name);

/**
 * opengl_has_features:  Return whether the given OpenGL features are all
 * available.
 *
 * [Parameters]
 *     features: Features to check.
 * [Return value]
 *     True if all of the specified features are available, false otherwise.
 */
extern PURE_FUNCTION int opengl_has_features(int features);

/**
 * opengl_has_formats:  Return whether the given OpenGL formats are all
 * supported.
 *
 * [Parameters]
 *     formats: Formats to check.
 * [Return value]
 *     True if all of the specified formats are supported, false otherwise.
 */
extern PURE_FUNCTION int opengl_has_formats(int formats);

/**
 * opengl_set_delete_buffer_size:  Set the size of the buffer used to hold
 * objects pending deletion.  The buffer will never change size; if more
 * objects are deleted without being flushed than can fit in the array,
 * the object which overflows the array will trigger an immediate flush of
 * all pending deletions.
 *
 * If size == 0, the buffer will instead be dynamically allocated and
 * expanded as necessary to hold objects pending deletion.  This is the
 * default behavior.
 *
 * Calling this function will flush all pending deletions and free the
 * existing delete buffer, if any.
 *
 * This function does nothing if the OPENGL_FEATURE_DELAYED_DELETE feature
 * flag was not passed to opengl_init().
 *
 * [Parameters]
 *     size: Maximum number of elements to store, or zero to use dynamic
 *         allocation.
 * [Return value]
 *     True on success, false on insufficient memory.
 */
int opengl_set_delete_buffer_size(int size);

/**
 * opengl_start_frame:  Perform OpenGL-specific setup at the beginning of
 * the frame.  (There is no corresponding opengl_finish_frame(), as it is
 * not currently required.)
 */
extern void opengl_start_frame(void);

/**
 * opengl_sync:  Wait for background rendering to complete.
 */
extern void opengl_sync(void);

/**
 * opengl_free_dead_resources:  Free all resources which were destroyed
 * since the last call to this function.  If the DELAYED_DELETE feature
 * flag is passed to opengl_init(), the system-dependent code must
 * periodically call this function (typically at the beginning of a frame)
 * to prevent resources from leaking.
 *
 * If the DELAYED_DELETE feature flag was not passed to opengl_init(),
 * this function does nothing.
 *
 * [Parameters]
 *     also_array: True to also free any memory allocated dynamically for
 *         storing pending deletions; false to leave such memory allocated.
 */
extern void opengl_free_dead_resources(int also_array);

/**
 * opengl_save_state:  Save all current OpenGL state (including pixel data
 * for textures, but not the contents of user-created framebuffers) in
 * local memory, so it can be later restored by opengl_restore_state().
 * This should be called before an operation that would invalidate the
 * current OpenGL context, such as closing the window on a multi-window
 * platform.
 *
 * This function can be called on OpenGL ES systems, but it will fail to
 * preserve texture and primitive data.
 */
extern void opengl_save_state(void);

/**
 * opengl_restore_state:  Restore OpenGL state saved by opengl_save_state().
 * This should be called after the OpenGL context has been restored.
 */
extern void opengl_restore_state(void);


/******** shader.c ********/

/**
 * opengl_set_shader_callbacks:  Set the callback functions to be called to
 * generate key values or source code for a shader.  Calling this function
 * invalidates and destroys any shaders which have already been created.
 *
 * When choosing an initial hash table size, keep in mind that dynamically
 * resizing the hash table can be an expensive operation; if memory permits
 * (32 bytes per table entry), set the initial table size large enough to
 * hold all possible permutations of shader parameters.
 *
 * [Parameters]
 *     source_callback: Source code generation function to call (NULL for
 *         the default function).
 *     key_callback: Key generation function to call (mandatory if
 *         source_callback != NULL; ignored otherwise).
 *     hash_table_size: Initial size of the shader hash table.
 *     dynamic_resize: True to allow dynamic resizing of the hash table,
 *         false to flush old shaders when the table is full.
 */
extern void opengl_set_shader_callbacks(
    ShaderSourceCallback *source_callback, ShaderKeyCallback *key_callback,
    int hash_table_size, int dynamic_resize);


/******** texture.c ********/

/**
 * opengl_texture_create_external:  Create a texture for use with the
 * OpenGL ES texture target GL_TEXTURE_EXTERNAL_OES.
 *
 * [Parameters]
 *     mem_flags: Memory allocation flags (MEM_ALLOC_*).
 * [Return value]
 *     Newly created texture, or NULL on error.
 */
extern SysTexture *opengl_texture_create_external(int mem_flags);

/**
 * opengl_texture_id:  Return the OpenGL texture ID for the given texture.
 *
 * [Parameters]
 *     texture: SysTexture object.
 * [Return value]
 *     OpenGL texture ID.
 */
extern int opengl_texture_id(const SysTexture *texture);

/*************************************************************************/
/*************************** Test control data ***************************/
/*************************************************************************/

#ifdef INCLUDE_TESTS

/**
 * TEST_opengl_force_feature_flags, TEST_opengl_force_feature_mask:  These
 * variables operate in tandem to force certain OpenGL feature flags on or
 * off regardless of what the system-specific code requests.  When
 * opengl_init() is called, the features parameter is modified as follows:
 *
 *     features = ((features & ~TEST_opengl_force_feature_mask)
 *                 | TEST_opengl_force_feature_flags);
 *
 * Either of the following cases will trigger an assertion failure:
 *
 * - (TEST_opengl_force_feature_flags & ~TEST_opengl_force_feature_mask) != 0
 *   i.e., trying to set a flag bit not included in the mask
 *
 * - ((features & ..._feature_mask) ^ ..._feature_flags) != ..._feature_mask
 *   i.e., trying to set or clear a bit that is already in the target state
 */
extern int TEST_opengl_force_feature_flags;
extern int TEST_opengl_force_feature_mask;

/**
 * TEST_opengl_always_convert_texture_data:  If true, sys_texture_create()
 * will always use a pixel format conversion function when one exists for
 * the input texture data format.
 */
extern uint8_t TEST_opengl_always_convert_texture_data;

#endif

/*************************************************************************/
/*************************************************************************/

#endif  // SIL_SRC_SYSDEP_OPENGL_OPENGL_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
