/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/opengl/texture.c: Texture manipulation functionality for
 * OpenGL-based ports.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/graphics.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/opengl/internal.h"
#include "src/sysdep/opengl/opengl.h"
#include "src/texture.h"
#include "src/utility/dds.h"
#include "src/utility/memory.h"
#include "src/utility/png.h"
#include "src/utility/tex-file.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Global texture list (used when restoring textures after device loss). */
static SysTexture *all_textures;

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * read_texture_via_framebuffer:  Read the given texture's data by
 * attaching the texture to a framebuffer and calling glReadPixels() on
 * the framebuffer.
 *
 * [Parameters]
 *     texture: Texture to read from.
 *     x, y, w, h: Coordinates of region to read, in pixels.
 *     buffer: Output pixel buffer (RGBA format).
 * [Return value]
 *     True on success, false on error.
 */
static int read_texture_via_framebuffer(
    SysTexture *texture, int x, int y, int w, int h, void *buffer);

/**
 * update_texture:  Update the OpenGL texture object for the given texture
 * using the data in the given pixel buffer, generating mipmaps if the
 * texture's auto_mipmaps flag is set.
 *
 * The input pixel buffer is assumed to have been allocated with
 * mem_alloc(), and will be freed with mem_free() before return.  While
 * this does complicate the caller's logic, it reduces peak memory usage
 * by releasing the local copy of the higher-resolution data before
 * sending the lower-resolution data to the GL.
 *
 * [Parameters]
 *     texture: Texture to update.
 *     pixels: Pixel data to store; will be freed before return.
 */
static void update_texture(SysTexture *texture, uint8_t *pixels);

/**
 * generate_mipmaps:  Generate mipmaps for the given texture (down to size
 * 1x1) and register them with the OpenGL driver.  The texture ID is
 * assumed to be already bound to GL_TEXTURE_2D.
 *
 * [Parameters]
 *     pixels: Pixel data buffer, in RGBA format; will be freed before return.
 *     offset: Byte offset from "pixels" to start of pixel data.
 *     width: Width of current texture image, in pixels.
 *     height: Height of current texture image, in pixels.
 *     level: Mipmap level of current texture image (0 = primary image).
 */
static void generate_mipmaps(uint8_t *pixels, int32_t offset,
                             int width, int height, int level);

/**
 * ConvertFunc, convert_*:  Function type and individual functions for
 * converting between different pixel formats.  If the graphics hardware
 * doesn't support a particular uncompressed format, we use one of these
 * functions to convert to a format that is supported.
 *
 * The actual conversion functions are named "convert_<from>_<to>".  Bit
 * patterns are labeled following the OpenGL style, thus "bgra1555rev"
 * refers to format GL_BGRA, data type GL_UNSIGNED_SHORT_1_5_5_5_REV.
 *
 * [Parameters]
 *     dest: Output data pointer.
 *     src: Input data pointer (may be equal to dest if the data sizes are
 *         the same).
 *     len: Number of pixels to process.
 */
typedef void ConvertFunc(void *dest, const void *src, size_t len);
static void convert_bgra8888_rgba8888(void *dest, const void *src, size_t len);
static void convert_565rev_565(void *dest, const void *src, size_t len);
static void convert_rgba1555rev_rgba5551(void *dest, const void *src,
                                         size_t len);
static void convert_bgra1555rev_rgba5551(void *dest, const void *src,
                                         size_t len);
static void convert_rgba4444rev_rgba4444(void *dest, const void *src,
                                         size_t len);
static void convert_bgra4444rev_rgba4444(void *dest, const void *src,
                                         size_t len);
static void convert_a8_rgba8888(void *dest, const void *src, size_t len);

/*************************************************************************/
/*************** Interface: Texture creation and deletion ****************/
/*************************************************************************/

SysTexture *sys_texture_create(
    int width, int height, TextureFormat data_format, int num_levels,
    void *data, int stride, const int32_t *level_offsets,
    const int32_t *level_sizes, int mipmaps, int mem_flags, int reuse)
{
    /* Determine the GL format parameters for the texture. */

    GLenum gl_format = 0, gl_type = 0;
    int in_bpp = 0, bpp = 0, is_palette = 0, is_compressed = 0;
    ConvertFunc *convert_func = NULL;

#ifdef INCLUDE_TESTS
    const int force_convert = TEST_opengl_always_convert_texture_data;
#else
    const int force_convert = 0;
#endif

    switch (data_format) {

      case TEX_FORMAT_RGBA8888:
        gl_format = GL_RGBA;
        gl_type = GL_UNSIGNED_BYTE;
        bpp = 32;
        break;

      case TEX_FORMAT_RGB565:
        gl_format = GL_RGB;
        /* _REV because we label components from the low bits, but OpenGL
         * labels from the high bits. */
        if (opengl_has_formats(OPENGL_FORMAT_BITREV) && !force_convert) {
            gl_type = GL_UNSIGNED_SHORT_5_6_5_REV;
        } else {
            convert_func = convert_565rev_565;
            gl_type = GL_UNSIGNED_SHORT_5_6_5;
        }
        bpp = 16;
        break;

      case TEX_FORMAT_RGBA5551:
        gl_format = GL_RGBA;
        if (opengl_has_formats(OPENGL_FORMAT_BITREV) && !force_convert) {
            gl_type = GL_UNSIGNED_SHORT_1_5_5_5_REV;
        } else {
            convert_func = convert_rgba1555rev_rgba5551;
            gl_type = GL_UNSIGNED_SHORT_5_5_5_1;
        }
        bpp = 16;
        break;

      case TEX_FORMAT_RGBA4444:
        gl_format = GL_RGBA;
        if (opengl_has_formats(OPENGL_FORMAT_BITREV) && !force_convert) {
            gl_type = GL_UNSIGNED_SHORT_4_4_4_4_REV;
        } else {
            convert_func = convert_rgba4444rev_rgba4444;
            gl_type = GL_UNSIGNED_SHORT_4_4_4_4;
        }
        bpp = 16;
        break;

      case TEX_FORMAT_BGRA8888:
        if (opengl_has_formats(OPENGL_FORMAT_BGRA) && !force_convert) {
            gl_format = GL_BGRA;
        } else {
            convert_func = convert_bgra8888_rgba8888;
            gl_format = GL_RGBA;
        }
        gl_type = GL_UNSIGNED_BYTE;
        bpp = 32;
        break;

      case TEX_FORMAT_BGR565:
        gl_format = GL_RGB;
        gl_type = GL_UNSIGNED_SHORT_5_6_5;
        bpp = 16;
        break;

      case TEX_FORMAT_BGRA5551:
        if (opengl_has_formats(OPENGL_FORMAT_BGRA | OPENGL_FORMAT_BITREV)
         && !force_convert) {
            gl_format = GL_BGRA;
            gl_type = GL_UNSIGNED_SHORT_1_5_5_5_REV;
        } else {
            convert_func = convert_bgra1555rev_rgba5551;
            gl_format = GL_RGBA;
            gl_type = GL_UNSIGNED_SHORT_5_5_5_1;
        }
        bpp = 16;
        break;

      case TEX_FORMAT_BGRA4444:
        if (opengl_has_formats(OPENGL_FORMAT_BGRA | OPENGL_FORMAT_BITREV)
         && !force_convert) {
            gl_format = GL_BGRA;
            gl_type = GL_UNSIGNED_SHORT_4_4_4_4_REV;
        } else {
            convert_func = convert_bgra4444rev_rgba4444;
            gl_format = GL_RGBA;
            gl_type = GL_UNSIGNED_SHORT_4_4_4_4;
        }
        bpp = 16;
        break;

      case TEX_FORMAT_A8:
        if (opengl_has_formats(OPENGL_FORMAT_ALPHA) && !force_convert) {
            gl_format = GL_ALPHA;
            gl_type = GL_UNSIGNED_BYTE;
            bpp = 8;
        } else {
            convert_func = convert_a8_rgba8888;
            gl_format = GL_RGBA;
            gl_type = GL_UNSIGNED_BYTE;
            in_bpp = 8;
            bpp = 32;
        }
        break;

      case TEX_FORMAT_PSP_RGBA8888:
      case TEX_FORMAT_PSP_RGB565:
      case TEX_FORMAT_PSP_RGBA5551:
      case TEX_FORMAT_PSP_RGBA4444:
      case TEX_FORMAT_PSP_PALETTE8_RGBA8888:
      case TEX_FORMAT_PSP_A8:
      case TEX_FORMAT_PSP_RGBA8888_SWIZZLED:
      case TEX_FORMAT_PSP_RGB565_SWIZZLED:
      case TEX_FORMAT_PSP_RGBA5551_SWIZZLED:
      case TEX_FORMAT_PSP_RGBA4444_SWIZZLED:
      case TEX_FORMAT_PSP_PALETTE8_RGBA8888_SWIZZLED:
      case TEX_FORMAT_PSP_A8_SWIZZLED:
        DLOG("Pixel format %u unsupported", data_format);
        goto error_return;

      case TEX_FORMAT_PALETTE8_RGBA8888:
        gl_format = GL_RGBA;
        gl_type = GL_UNSIGNED_BYTE;
        bpp = 8;
        is_palette = 1;
        break;

      case TEX_FORMAT_S3TC_DXT1:
        gl_format = gl_type = GL_COMPRESSED_RGB_S3TC_DXT1_EXT;
        bpp = 4;
        is_compressed = 1;
        break;

      case TEX_FORMAT_S3TC_DXT3:
        gl_format = gl_type = GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
        bpp = 8;
        is_compressed = 1;
        break;

      case TEX_FORMAT_S3TC_DXT5:
        gl_format = gl_type = GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
        bpp = 8;
        is_compressed = 1;
        break;

      case TEX_FORMAT_PVRTC2_RGBA:
        gl_format = gl_type = GL_COMPRESSED_RGBA_PVRTC_2BPPV1_IMG;
        bpp = 2;
        is_compressed = 1;
        break;

      case TEX_FORMAT_PVRTC2_RGB:
        gl_format = gl_type = GL_COMPRESSED_RGB_PVRTC_2BPPV1_IMG;
        bpp = 2;
        is_compressed = 1;
        break;

      case TEX_FORMAT_PVRTC4_RGBA:
        gl_format = gl_type = GL_COMPRESSED_RGBA_PVRTC_4BPPV1_IMG;
        bpp = 4;
        is_compressed = 1;
        break;

      case TEX_FORMAT_PVRTC4_RGB:
        gl_format = gl_type = GL_COMPRESSED_RGB_PVRTC_4BPPV1_IMG;
        bpp = 4;
        is_compressed = 1;
        break;

    }  // switch (data_format)

    /* Use this test instead of a default case so we get compiler warnings
     * if we forget to handle any data types. */
    if (UNLIKELY(bpp == 0)) {
        DLOG("Pixel format %u unknown", data_format);
        goto error_return;
    }

    if (in_bpp == 0) {
        in_bpp = bpp;
    }

    const GLenum gl_components =
        (gl_format==GL_ALPHA ? GL_ALPHA :
#ifdef SIL_OPENGL_ES
         /* GLES and plain GL differ in the expected value here for BGRA
          * format.  To make things worse, Apple deviates from the GLES
          * standard (EXT_texture_format_BGRA8888) and uses the non-ES
          * style. */
         gl_format==GL_BGRA && !opengl_has_extension("GL_APPLE_texture_format_BGRA8888") ? GL_BGRA :
#endif
         gl_format==GL_RGB || gl_format==GL_COMPRESSED_RGB_PVRTC_2BPPV1_IMG
                           || gl_format==GL_COMPRESSED_RGB_PVRTC_4BPPV1_IMG
         ? GL_RGB : GL_RGBA);

    if (UNLIKELY(is_compressed
                 && !opengl_has_features(OPENGL_FEATURE_COMPRESSEDTEXIMAGE))) {
        DLOG("glCompressedTexImage2D() unavailable, can't use format %u",
             data_format);
        goto error_return;
    }

    /* Allocate and set up the SysTexture structure. */

    SysTexture *texture = debug_mem_alloc(
        sizeof(*texture), 0, mem_flags & ~MEM_ALLOC_CLEAR,
        __FILE__, __LINE__, MEM_INFO_TEXTURE);
    if (UNLIKELY(!texture)) {
        DLOG("Failed to allocate SysTexture");
        goto error_return;
    }

    opengl_clear_error();
    texture->id = 0;
    glGenTextures(1, &texture->id);
    if (UNLIKELY(!texture->id)) {
        DLOG("Failed to generate an OpenGL texture ID: 0x%04X", glGetError());
        goto error_free_texture;
    }

    texture->width      = width;
    texture->height     = height;
    texture->color_type = (gl_format==GL_ALPHA ? TEXCOLOR_A : TEXCOLOR_RGBA);
    texture->repeat_u   = 1;
    texture->repeat_v   = 1;
    texture->antialias  = 1;
    texture->empty      = (num_levels == 0);
    texture->readable   = (!is_compressed && gl_format == GL_RGBA);
    texture->lock_buf   = NULL;
    texture->save_buf   = NULL;

    if (mipmaps) {
        /* Core OpenGL (non-ES) allows mipmaps on textures of any size;
         * OpenGL ES only allows them for power-of-two-sized textures.
         * Our custom generate_mipmaps() routine also only handles
         * power-of-two textures, and is limited to 32bpp textures. */
        if (opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)) {
#ifdef SIL_OPENGL_ES
            texture->auto_mipmaps = ((width  & (width -1)) == 0
                                  && (height & (height-1)) == 0);
#else
            texture->auto_mipmaps = 1;
#endif
        } else {
            texture->auto_mipmaps = (in_bpp == 32
                                  && (width  & (width -1)) == 0
                                  && (height & (height-1)) == 0);
        }
    } else {
        texture->auto_mipmaps = 0;
    }
    if (texture->auto_mipmaps && num_levels > 1) {
        num_levels = 1; // Ignore provided mipmap data since we'll generate it.
    }
    texture->has_mipmaps = texture->auto_mipmaps || (num_levels > 1);

    /* Load the texture data (if any). */

    uint32_t palette[256];

    glBindTexture(GL_TEXTURE_2D, texture->id);
    /* The minification filter defaults to NEAREST_MIPMAP_LINEAR, which
     * in GLES prevents the creation of non-power-of-two textures, so
     * temporarily change it here.  We'll set it for real later on. */
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

    if (num_levels == 0) {
        glTexImage2D(GL_TEXTURE_2D, 0, gl_components, width, height, 0,
                     GL_RGBA, GL_UNSIGNED_BYTE, NULL);
        if (texture->has_mipmaps) {
            int level = 0, w = texture->width, h = texture->height;
            while (w > 1 || h > 1) {
                level++;
                w = (w+1) / 2;
                h = (h+1) / 2;
                glTexImage2D(GL_TEXTURE_2D, level, GL_RGBA, w, h, 0,
                             GL_RGBA, GL_UNSIGNED_BYTE, NULL);
            }
        }
    } else {
        for (int level = 0; level < num_levels; level++) {

            const int level_w = lbound(width >> level, 1);
            const int level_h = lbound(height >> level, 1);
            const int level_s = lbound(stride >> level, 1);
            uint8_t *level_data = (uint8_t *)data + level_offsets[level];
            int32_t level_size = level_sizes[level];

            if (is_compressed) {
                glCompressedTexImage2D(GL_TEXTURE_2D, level, gl_format,
                                       level_w, level_h, 0, level_size,
                                       level_data);

            } else if (is_palette) {
                if (level == 0) {
                    memcpy(palette, level_data, sizeof(palette));
                    level_data += sizeof(palette);
                    level_size -= sizeof(palette);
                }
                uint32_t *temp_data =
                    mem_alloc(level_size*4, 0, MEM_ALLOC_TEMP);
                if (!temp_data) {
                    DLOG("Failed to get memory for level %d palette fixup"
                         " (%d bytes)", level, level_size*4);
                    goto error_delete_gltex;
                }
                for (int32_t i = 0; i < level_size; i++) {
                    temp_data[i] = palette[level_data[i]];
                }
                glTexImage2D(GL_TEXTURE_2D, level, gl_components, level_w,
                             level_h, 0, gl_format, gl_type, temp_data);
                mem_free(temp_data);

            } else if ((level_s != level_w || convert_func)
                       && (!reuse || bpp != in_bpp)) {
                const int32_t temp_size = (level_w * level_h * bpp) / 8;
                /* Ensure 4-byte alignment for OpenGL (see below). */
                uint8_t *temp_data = mem_alloc(temp_size, 4, MEM_ALLOC_TEMP);
                if (!temp_data) {
                    DLOG("Failed to get memory for level %d data conversion"
                         " (%d bytes)", level, temp_size);
                    goto error_delete_gltex;
                }
                if (level_s != level_w) {
                    if (convert_func) {
                        for (int y = 0; y < level_h; y++) {
                            (*convert_func)(temp_data + y*((level_w*bpp)/8),
                                            level_data + y*((level_s*in_bpp)/8),
                                            level_w);
                        }
                    } else {
                        for (int y = 0; y < level_h; y++) {
                            memcpy(temp_data + y*((level_w*bpp)/8),
                                   level_data + y*((level_s*in_bpp)/8),
                                   (level_w*bpp)/8);
                        }
                    }
                } else {  // level_s == level_w && convert_func
                    (*convert_func)(temp_data, level_data, level_w * level_h);
                }
                /* The default data alignment is 4 bytes, so we need to
                 * adjust if the width for this image level is not a
                 * multiple of that (we ensure data buffer alignment above).
                 * Otherwise, we leave the alignment alone so as not to
                 * trigger the use of a slower copy algorithm in the
                 * graphics driver.  We assume each pixel is aligned on
                 * a natural boundary. */
                const int is_unaligned = ((bpp == 8 && level_w % 4 != 0)
                                          || (bpp == 16 && level_w % 2 != 0));
                if (is_unaligned) {
                    glPixelStorei(GL_UNPACK_ALIGNMENT, bpp/8);
                }
                glTexImage2D(GL_TEXTURE_2D, level, gl_components, level_w,
                             level_h, 0, gl_format, gl_type, temp_data);
                if (is_unaligned) {
                    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
                }
                mem_free(temp_data);

            } else {
                if (level_s != level_w) {
                    /* We must be reusing the data, so we can write over the
                     * input buffer. */
                    if (convert_func) {
                        for (int y = 0; y < level_h; y++) {
                            (*convert_func)(level_data + y*((level_w*bpp)/8),
                                            level_data + y*((level_s*bpp)/8),
                                            level_w);
                        }
                    } else {
                        for (int y = 1; y < level_h; y++) {
                            memcpy(level_data + y*((level_w*bpp)/8),
                                   level_data + y*((level_s*bpp)/8),
                                   (level_w*bpp)/8);
                        }
                    }
                } else if (convert_func) {
                    /* Again, we must be reusing the data. */
                    (*convert_func)(level_data, level_data, level_w * level_h);
                }
                const int is_unaligned =
                    ((bpp == 8 && ((uintptr_t)level_data % 4 != 0
                                   || level_w % 4 != 0))
                     || (bpp == 16 && ((uintptr_t)level_data % 2 != 0
                                       || level_w % 2 != 0)));
                if (is_unaligned) {
                    glPixelStorei(GL_UNPACK_ALIGNMENT, bpp/8);
                }
                glTexImage2D(GL_TEXTURE_2D, level, gl_components, level_w,
                             level_h, 0, gl_format, gl_type, level_data);
                if (is_unaligned) {
                    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
                }

            }
        }  // for (int level = 0; level < num_levels; level++) {
    }  // if (num_levels == 0)

    if (texture->auto_mipmaps && num_levels > 0) {
        if (opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)) {
            glGenerateMipmap(GL_TEXTURE_2D);
        } else {
            int32_t offset;
            if (reuse) {
                offset = level_offsets[0];
            } else {
                void *new_data = mem_alloc(level_sizes[0], 0, MEM_ALLOC_TEMP);
                if (!new_data) {
                    DLOG("No memory for mipmap generation, skipping");
                    texture->has_mipmaps = 0;
                    texture->auto_mipmaps = 0;
                    goto no_mipmaps;
                } else {
                    memcpy(new_data, (const uint8_t *)data + level_offsets[0],
                           level_sizes[0]);
                    data = new_data;
                }
                offset = 0;
            }
            generate_mipmaps(data, offset, texture->width, texture->height, 0);
            data = NULL;  // Freed by generate_mipmaps().
        }
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                        GL_LINEAR_MIPMAP_LINEAR);
    }
    if (texture->has_mipmaps) {
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                        GL_LINEAR_MIPMAP_LINEAR);
    } else {
      no_mipmaps:
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    }
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    if (opengl_current_texture_target == GL_TEXTURE_2D) {
        glBindTexture(GL_TEXTURE_2D, opengl_current_texture_id);
    } else {
        glBindTexture(GL_TEXTURE_2D, 0);
    }
    const int error = glGetError();
    if (UNLIKELY(error != GL_NO_ERROR)) {
        DLOG("Failed to initialize texture data (%dx%d): 0x%04X",
             width, height, error);
        goto error_delete_gltex;
    }

    /* Link the texture into the global list and return. */

    texture->next = all_textures;
    if (all_textures) {
        all_textures->prev = texture;
    }
    texture->prev = NULL;
    all_textures = texture;

    if (reuse) {
        mem_free(data);
    }
    return texture;

    /* Error handling. */

  error_delete_gltex:
    opengl_delete_texture(texture->id);
  error_free_texture:
    mem_free(texture);
  error_return:
    if (reuse) {
        mem_free(data);
    }
    return NULL;
}

/*-----------------------------------------------------------------------*/

void sys_texture_destroy(SysTexture *texture)
{
    if (texture->next) {
        texture->next->prev = texture->prev;
    }
    if (texture->prev) {
        texture->prev->next = texture->next;
    } else {
        all_textures = texture->next;
    }

    if (opengl_current_texture_id == texture->id) {
        glBindTexture(opengl_current_texture_target, 0);
        opengl_current_texture_id = 0;
    }
    if (opengl_current_texture == texture) {
        opengl_current_texture = NULL;
    }

    opengl_delete_texture(texture->id);
    mem_free(texture->lock_buf);
    mem_free(texture->save_buf);
    mem_free(texture);
}

/*************************************************************************/
/*************** Interface: Texture information retrieval ****************/
/*************************************************************************/

int sys_texture_width(SysTexture *texture)
{
    return texture->width;
}

/*-----------------------------------------------------------------------*/

int sys_texture_height(SysTexture *texture)
{
    return texture->height;
}

/*-----------------------------------------------------------------------*/

int sys_texture_has_mipmaps(SysTexture *texture)
{
    return texture->has_mipmaps;
}

/*************************************************************************/
/****************** Interface: Pixel data manipulation *******************/
/*************************************************************************/

int sys_texture_grab(SysTexture *texture, int x, int y, int w, int h,
                     UNUSED int readable)
{
    if ((w == texture->width && h == texture->height)
     && !(texture->auto_mipmaps
          && !opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)))
    {
        /* Fast case: perform the copy entirely within GL. */
        glBindTexture(GL_TEXTURE_2D, texture->id);
        glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, x, y, w, h, 0);
        if (texture->auto_mipmaps) {
            glGenerateMipmap(GL_TEXTURE_2D);
            texture->has_mipmaps = 1;
        } else {
            texture->has_mipmaps = 0;
        }
        if (opengl_current_texture_target == GL_TEXTURE_2D) {
            glBindTexture(GL_TEXTURE_2D, opengl_current_texture_id);
        } else {
            glBindTexture(GL_TEXTURE_2D, 0);
        }
        texture->empty = 0;
        texture->color_type = TEXCOLOR_RGB;
        texture->readable = 1;
        return 1;
    }

    /* Slow case: partial copy or manual mipmaps needed. */
    uint8_t *pixels = sys_texture_lock(texture, SYS_TEXTURE_LOCK_DISCARD,
                                       0, 0, w, h);
    if (UNLIKELY(!pixels)) {
        DLOG("Failed to lock texture, can't grab");
        return 0;
    }
    glReadPixels(x, y, w, h, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
    for (int i = 0; i < w*h; i++) {
        pixels[i*4+3] = 0xFF;
    }
    sys_texture_unlock(texture, 1);
    return 1;
}

/*-----------------------------------------------------------------------*/

uint8_t *sys_texture_lock(SysTexture *texture, SysTextureLockMode lock_mode,
                          int x, int y, int w, int h)
{
    const uint32_t size = w * h * 4;
    texture->lock_buf = mem_alloc(size, SIL_OPENGL_TEXTURE_BUFFER_ALIGNMENT,
                                  MEM_ALLOC_TEMP);
    if (UNLIKELY(!texture->lock_buf)) {
        DLOG("lock(%p): Failed to get a lock buffer (%u bytes)",
             texture, size);
        return NULL;
    }

    if (lock_mode == SYS_TEXTURE_LOCK_DISCARD) {
        /* Nothing else to do. */

    } else if (texture->empty) {
        mem_clear(texture->lock_buf, size);

    } else if (opengl_has_features(OPENGL_FEATURE_GETTEXIMAGE)) {
        uint32_t *texture_buf;
        if (w == texture->width && h == texture->height) {
            texture_buf = (uint32_t *)texture->lock_buf;
        } else {
            /* There's no glGetTexSubImage(), so we'll need to fetch the
             * entire texture data and copy out of it. */
            const uint32_t full_size = texture->width * texture->height * 4;
            texture_buf = mem_alloc(full_size, 0, MEM_ALLOC_TEMP);
            if (UNLIKELY(!texture_buf)) {
                DLOG("lock(%p): Failed to get a texture image buffer (%u"
                     " bytes)", texture, full_size);
                mem_free(texture->lock_buf);
                texture->lock_buf = NULL;
                return NULL;
            }
        }

        opengl_clear_error();
        glBindTexture(GL_TEXTURE_2D, texture->id);
        glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, texture_buf);
        if (opengl_current_texture_target == GL_TEXTURE_2D) {
            glBindTexture(GL_TEXTURE_2D, opengl_current_texture_id);
        } else {
            glBindTexture(GL_TEXTURE_2D, 0);
        }
        const int error = glGetError();
        if (UNLIKELY(error != GL_NO_ERROR)) {
            DLOG("lock(%p): Failed to retrieve texture data (%dx%d): 0x%04X",
                 texture, texture->width, texture->height, error);
            if (texture_buf != texture->lock_buf) {
                mem_free(texture_buf);
            }
            mem_free(texture->lock_buf);
            texture->lock_buf = NULL;
            return NULL;
        }

        if (texture_buf != texture->lock_buf) {
            const uint32_t *src = texture_buf + y*texture->width + x;
            uint32_t *dest = (uint32_t *)texture->lock_buf;
            if (w == texture->width) {
                memcpy(dest, src, h * texture->width * 4);
            } else {
                for (int yy = 0; yy < h; yy++, src += texture->width, dest += w) {
                    memcpy(dest, src, w*4);
                }
            }
            mem_free(texture_buf);
        }

        if (texture->color_type == TEXCOLOR_A) {
            /* OpenGL gives us RGB=(0,0,0), but we want RGB=(255,255,255). */
            uint8_t *pixels = texture->lock_buf;
            for (int i = 0; i < w*h; i++) {
                pixels[i*4+0] = 255;
                pixels[i*4+1] = 255;
                pixels[i*4+2] = 255;
            }
        }

    } else if (texture->readable) {
        if (!read_texture_via_framebuffer(texture, x, y, w, h,
                                          texture->lock_buf)) {
            mem_free(texture->lock_buf);
            texture->lock_buf = NULL;
            return NULL;
        }

    } else {
        DLOG("Can't read texture data on this platform");
        mem_free(texture->lock_buf);
        texture->lock_buf = NULL;
        return NULL;
    }

    texture->lock_mode = lock_mode;
    texture->lock_x    = x;
    texture->lock_y    = y;
    texture->lock_w    = w;
    texture->lock_h    = h;
    return texture->lock_buf;
}

/*-----------------------------------------------------------------------*/

void sys_texture_unlock(SysTexture *texture, int update)
{
    if (update) {

        if (texture->lock_w == texture->width
         && texture->lock_h == texture->height) {
            update_texture(texture, texture->lock_buf);
            texture->lock_buf = NULL;

        } else if ((!texture->auto_mipmaps
                    || opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP))
                   && texture->color_type == TEXCOLOR_RGBA) {
            glBindTexture(GL_TEXTURE_2D, texture->id);
            glTexSubImage2D(GL_TEXTURE_2D, 0,
                            texture->lock_x, texture->lock_y,
                            texture->lock_w, texture->lock_h,
                            GL_RGBA, GL_UNSIGNED_BYTE, texture->lock_buf);
            /* Free the old buffer early to reduce memory pressure. */
            mem_free(texture->lock_buf);
            texture->lock_buf = NULL;
            texture->empty = 0;
            if (texture->auto_mipmaps) {
                glGenerateMipmap(GL_TEXTURE_2D);
                texture->has_mipmaps = 1;
            } else {
                texture->has_mipmaps = 0;
            }
            if (opengl_current_texture_target == GL_TEXTURE_2D) {
                glBindTexture(GL_TEXTURE_2D, opengl_current_texture_id);
            } else {
                glBindTexture(GL_TEXTURE_2D, 0);
            }

        } else if (opengl_has_features(OPENGL_FEATURE_GETTEXIMAGE)) {
            const uint32_t size = texture->width * texture->height * 4;
            uint32_t *buf = mem_alloc(size, 0, MEM_ALLOC_TEMP);
            if (UNLIKELY(!buf)) {
                DLOG("unlock(%p): Failed to get a buffer (%u bytes),"
                     " skipping partial update", texture, size);
            } else {
                glBindTexture(GL_TEXTURE_2D, texture->id);
                glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, buf);
                if (opengl_current_texture_target == GL_TEXTURE_2D) {
                    glBindTexture(GL_TEXTURE_2D, opengl_current_texture_id);
                } else {
                    glBindTexture(GL_TEXTURE_2D, 0);
                }
                const uint32_t *src = (const uint32_t *)texture->lock_buf;
                uint32_t *dest =
                    buf + texture->lock_y*texture->width + texture->lock_x;
                if (texture->lock_w == texture->width) {
                    memcpy(dest, src, texture->lock_h * texture->width * 4);
                } else {
                    const int w = texture->lock_w, h = texture->lock_h;
                    for (int y = 0; y < h; y++, src+=w, dest+=texture->width) {
                        memcpy(dest, src, texture->lock_w * 4);
                    }
                }
                /* Free the old buffer early to reduce memory pressure. */
                mem_free(texture->lock_buf);
                texture->lock_buf = NULL;
                update_texture(texture, (void *)buf);
            }

        } else if (texture->readable) {
            const uint32_t full_size = texture->width * texture->height * 4;
            uint32_t *texture_buf = mem_alloc(full_size, 0, MEM_ALLOC_TEMP);
            if (UNLIKELY(!texture_buf)) {
                DLOG("unlock(%p): Failed to get a texture image buffer (%u"
                     " bytes)", texture, full_size);
            } else if (!read_texture_via_framebuffer(
                           texture, 0, 0, texture->width, texture->height,
                           texture_buf)) {
                DLOG("unlock(%p): Failed to read full texture data for"
                     " partial update", texture);
                mem_free(texture_buf);
            } else {
                const uint32_t *src = texture->lock_buf;
                uint32_t *dest = (texture_buf + texture->lock_y*texture->width
                                  + texture->lock_x);
                if (texture->lock_w == texture->width) {
                    memcpy(dest, src, texture->lock_h * texture->width * 4);
                } else {
                    const int w = texture->lock_w, h = texture->lock_h;
                    for (int y = 0; y < h; y++, src+=w, dest+=texture->width) {
                        memcpy(dest, src, w*4);
                    }
                }
                update_texture(texture, (uint8_t *)texture_buf);
            }

        } else {
            DLOG("unlock(%p): Unable to perform partial update on texture",
                 texture);
        }

    }  // if (update)

    mem_free(texture->lock_buf);  // In case we didn't free it above.
    texture->lock_buf = NULL;
}

/*-----------------------------------------------------------------------*/

void sys_texture_flush(UNUSED SysTexture *texture)
{
    /* Nothing to do for OpenGL.  We assume that if the caller has made
     * any direct OpenGL calls on the texture, the caller has also taken
     * responsibility for keeping the SysTexture structure and GL texture
     * state consistent. */
}

/*************************************************************************/
/******************** Interface: Rendering operations ********************/
/*************************************************************************/

void sys_texture_set_repeat(SysTexture *texture, int repeat_u, int repeat_v)
{
    int need_update = 0;
    if ((repeat_u != 0) != texture->repeat_u) {
        texture->repeat_u = (repeat_u != 0);
        need_update = 1;
    }
    if ((repeat_v != 0) != texture->repeat_v) {
        texture->repeat_v = (repeat_v != 0);
        need_update = 1;
    }
    if (need_update) {
        const GLenum target =
            (texture && texture->color_type == TEXCOLOR_EXTERNAL
             ? GL_TEXTURE_EXTERNAL_OES : GL_TEXTURE_2D);
        glBindTexture(target, texture->id);
        glTexParameteri(target, GL_TEXTURE_WRAP_S,
                        texture->repeat_u ? GL_REPEAT : GL_CLAMP_TO_EDGE);
        glTexParameteri(target, GL_TEXTURE_WRAP_T,
                        texture->repeat_v ? GL_REPEAT : GL_CLAMP_TO_EDGE);
        if (opengl_current_texture_target == target) {
            glBindTexture(target, opengl_current_texture_id);
        } else {
            glBindTexture(target, 0);
        }
    }
}

/*-----------------------------------------------------------------------*/

void sys_texture_set_antialias(SysTexture *texture, int on)
{
    const GLenum target =
        (texture && texture->color_type == TEXCOLOR_EXTERNAL
         ? GL_TEXTURE_EXTERNAL_OES : GL_TEXTURE_2D);

    if (on && !texture->antialias) {
        glBindTexture(target, texture->id);
        if (texture->has_mipmaps) {
            glTexParameteri(target, GL_TEXTURE_MIN_FILTER,
                            GL_LINEAR_MIPMAP_LINEAR);
        } else {
            glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        }
        glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        if (opengl_current_texture_target == target) {
            glBindTexture(target, opengl_current_texture_id);
        } else {
            glBindTexture(target, 0);
        }
        texture->antialias = 1;

    } else if (!on && texture->antialias) {
        glBindTexture(target, texture->id);
        if (texture->has_mipmaps) {
            glTexParameteri(target, GL_TEXTURE_MIN_FILTER,
                            GL_NEAREST_MIPMAP_NEAREST);
        } else {
            glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        }
        glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        if (opengl_current_texture_target == target) {
            glBindTexture(target, opengl_current_texture_id);
        } else {
            glBindTexture(target, 0);
        }
        texture->antialias = 0;
    }
}

/*-----------------------------------------------------------------------*/

void sys_texture_apply(SysTexture *texture)
{
    const GLenum target = (texture && texture->color_type == TEXCOLOR_EXTERNAL
                           ? GL_TEXTURE_EXTERNAL_OES : GL_TEXTURE_2D);
    opengl_apply_texture(target, texture ? texture->id : 0);
    opengl_texture_is_framebuffer = 0;
    opengl_current_texture = texture;
}

/*************************************************************************/
/*********************** Exported utility routines ***********************/
/*************************************************************************/

SysTexture *opengl_texture_create_external(int mem_flags)
{
    SysTexture *texture = debug_mem_alloc(
        sizeof(*texture), 0, mem_flags & ~MEM_ALLOC_CLEAR,
        __FILE__, __LINE__, MEM_INFO_TEXTURE);
    if (UNLIKELY(!texture)) {
        DLOG("Failed to allocate SysTexture");
        goto error_return;
    }

    opengl_clear_error();
    texture->id = 0;
    glGenTextures(1, &texture->id);
    if (UNLIKELY(!texture->id)) {
        DLOG("Failed to generate an OpenGL texture ID: 0x%04X", glGetError());
        goto error_free_texture;
    }

    texture->width      = 1;  // Placeholder values.
    texture->height     = 1;
    texture->color_type = TEXCOLOR_EXTERNAL;
    texture->repeat_u   = 0;
    texture->repeat_v   = 0;
    texture->antialias  = 1;
    texture->empty      = 0;
    texture->readable   = 0;
    texture->lock_buf   = NULL;
    texture->save_buf   = NULL;

    glBindTexture(GL_TEXTURE_EXTERNAL_OES, texture->id);
    glTexParameteri(GL_TEXTURE_EXTERNAL_OES, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_EXTERNAL_OES, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_EXTERNAL_OES,
                    GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_EXTERNAL_OES,
                    GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    if (opengl_current_texture
     && opengl_current_texture->color_type == TEXCOLOR_EXTERNAL) {
        glBindTexture(GL_TEXTURE_EXTERNAL_OES, opengl_current_texture->id);
    } else {
        glBindTexture(GL_TEXTURE_EXTERNAL_OES, 0);
    }
    const int error = glGetError();
    if (error != GL_NO_ERROR) {
        DLOG("Failed to initialize texture: 0x%04X", error);
        goto error_delete_gltex;
    }

    texture->next = all_textures;
    if (all_textures) {
        all_textures->prev = texture;
    }
    texture->prev = NULL;
    all_textures = texture;
    return texture;

  error_delete_gltex:
    opengl_delete_texture(texture->id);
  error_free_texture:
    mem_free(texture);
  error_return:
    return NULL;
}

/*-----------------------------------------------------------------------*/

int opengl_texture_id(const SysTexture *texture)
{
    PRECOND(texture != NULL, return 0);
    return texture->id;
}

/*************************************************************************/
/******************* Library-internal utility routines *******************/
/*************************************************************************/

void opengl_texture_save_all(void)
{
    opengl_apply_texture(0, 0);

    for (SysTexture *texture = all_textures; texture;
         opengl_delete_texture(texture->id), texture->id = 0,
             texture = texture->next)
    {
        if (texture->empty || texture->color_type == TEXCOLOR_EXTERNAL) {
            continue;
        }
        if (!opengl_has_features(OPENGL_FEATURE_GETTEXIMAGE)) {
            DLOG("save(%p): Can't read texture data, texture will be lost!",
                 texture);
            texture->save_buf = NULL;
            continue;
        }
        uint32_t size = texture->width * texture->height * 4;
        if (texture->has_mipmaps && !texture->auto_mipmaps) {
            int w = texture->width, h = texture->height;
            while (w > 1 || h > 1) {
                w = (w+1) / 2;
                h = (h+1) / 2;
                size += (w*h) * 4;
            }
        }
        texture->save_buf = mem_alloc(size, 0, MEM_ALLOC_TEMP);
        if (UNLIKELY(!texture->save_buf)) {
            DLOG("save(%p): Failed to get a buffer (%u bytes), texture will"
                 " be lost!", texture, size);
            continue;
        }
        opengl_clear_error();
        glBindTexture(GL_TEXTURE_2D, texture->id);
        glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE,
                      texture->save_buf);
        if (texture->has_mipmaps && !texture->auto_mipmaps) {
            uint32_t *dest = texture->save_buf;
            int level = 0, w = texture->width, h = texture->height;
            while (w > 1 || h > 1) {
                dest += w*h;
                level++;
                w = (w+1) / 2;
                h = (h+1) / 2;
                glGetTexImage(GL_TEXTURE_2D, level, GL_RGBA, GL_UNSIGNED_BYTE,
                              dest);
            }
        }
        const int error = glGetError();
        if (error != GL_NO_ERROR) {
            DLOG("save(%p): OpenGL error 0x%04X reading texture data,"
                 " texture will be lost!", texture, error);
            mem_free(texture->save_buf);
            texture->save_buf = NULL;
        }
    }
    glBindTexture(GL_TEXTURE_2D, 0);
}

/*-----------------------------------------------------------------------*/

void opengl_texture_restore_all(void)
{
    for (SysTexture *texture = all_textures; texture; texture = texture->next){
        ASSERT(texture->id == 0, continue);
        glGenTextures(1, &texture->id);
        if (UNLIKELY(!texture->id)) {
            DLOG("restore(%p): Failed to generate a new OpenGL texture",
                 texture);
            /* Free any saved pixel buffer to avoid memory leaks, but
             * otherwise the texture will be unusable. */
            mem_free(texture->save_buf);
            texture->save_buf = NULL;
            continue;
        }
        if (texture->color_type == TEXCOLOR_EXTERNAL) {
            /* Bind it once so it's marked as an external texture in the GL. */
            opengl_clear_error();
            glBindTexture(GL_TEXTURE_EXTERNAL_OES, texture->id);
            glBindTexture(GL_TEXTURE_EXTERNAL_OES, 0);
            const GLenum error = glGetError();
            if (error != GL_NO_ERROR) {
                DLOG("restore(%p): Failed to register as an external texture:"
                     " 0x%04X", texture, error);
            }
            continue;
        }
        opengl_clear_error();
        glBindTexture(GL_TEXTURE_2D, texture->id);
        /* Disable mipmaps for the texture load, as in sys_texture_create(). */
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        if (texture->save_buf) {
            if (texture->has_mipmaps && !texture->auto_mipmaps) {
                uint32_t *src = texture->save_buf;
                glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
                             texture->width, texture->height, 0,
                             GL_RGBA, GL_UNSIGNED_BYTE, src);
                int level = 0, w = texture->width, h = texture->height;
                while (w > 1 || h > 1) {
                    src += w*h;
                    level++;
                    w = (w+1) / 2;
                    h = (h+1) / 2;
                    glTexImage2D(GL_TEXTURE_2D, level, GL_RGBA, w, h, 0,
                                 GL_RGBA, GL_UNSIGNED_BYTE, src);
                }
                mem_free(texture->save_buf);
            } else {
                update_texture(texture, texture->save_buf);
                /* update_texture() resets the binding, so restore it. */
                glBindTexture(GL_TEXTURE_2D, texture->id);
            }
            texture->save_buf = NULL;
        } else {
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
                         texture->width, texture->height, 0,
                         GL_RGBA, GL_UNSIGNED_BYTE, NULL);
            if (texture->has_mipmaps) {
                int level = 0, w = texture->width, h = texture->height;
                while (w > 1 || h > 1) {
                    level++;
                    w = (w+1) / 2;
                    h = (h+1) / 2;
                    glTexImage2D(GL_TEXTURE_2D, level, GL_RGBA, w, h, 0,
                                 GL_RGBA, GL_UNSIGNED_BYTE, NULL);
                }
            }
            texture->empty = 1;
        }
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,
                        texture->repeat_u ? GL_REPEAT : GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,
                        texture->repeat_v ? GL_REPEAT : GL_CLAMP_TO_EDGE);
        if (texture->antialias) {
            if (texture->has_mipmaps) {
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                                GL_LINEAR_MIPMAP_LINEAR);
            } else {
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                                GL_LINEAR);
            }
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
                            GL_LINEAR);
        } else {
            if (texture->has_mipmaps) {
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                                GL_NEAREST_MIPMAP_NEAREST);
            } else {
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                                GL_NEAREST);
            }
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
                            GL_NEAREST);
        }
        const GLenum error = glGetError();
        if (error != GL_NO_ERROR) {
            DLOG("restore(%p): Failed to restore texture data: 0x%04X",
                 texture, error);
        }
    }
    glBindTexture(GL_TEXTURE_2D, 0);
    opengl_apply_texture(GL_TEXTURE_2D, 0);

    if (opengl_current_texture) {
        const GLenum target =
            (opengl_current_texture->color_type == TEXCOLOR_EXTERNAL
             ? GL_TEXTURE_EXTERNAL_OES : GL_TEXTURE_2D);
        opengl_apply_texture(target, opengl_current_texture->id);
    }
}

/*************************************************************************/
/****************** Local routines: Texture data access ******************/
/*************************************************************************/

static int read_texture_via_framebuffer(
    SysTexture *texture, int x, int y, int w, int h, void *buffer)
{
    PRECOND(texture != NULL, return 0);
    PRECOND(buffer != NULL, return 0);

    int result = 1;

    GLuint framebuffer = 0;
    glGenFramebuffers(1, &framebuffer);
    if (UNLIKELY(!framebuffer)) {
        DLOG("Failed to generate a framebuffer ID");
        result = 0;
    } else {
        glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);
        glBindTexture(GL_TEXTURE_2D, texture->id);
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
                               GL_TEXTURE_2D, texture->id, 0);
        const int status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
        if (status == GL_FRAMEBUFFER_COMPLETE) {
            glReadPixels(x, y, w, h, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
        } else {
            if (status == GL_FRAMEBUFFER_UNSUPPORTED) {
                DLOG("Framebuffer not supported by system (size %dx%d)", w, h);
            } else {
                DLOG("Unknown error 0x%X while setting up framebuffer (size"
                     " %dx%d)", status, w, h);
            }
            result = 0;
        }
        glDeleteFramebuffers(1, &framebuffer);
    }

    SysFramebuffer *current_framebuffer = opengl_current_framebuffer();
    if (current_framebuffer) {
        glBindFramebuffer(GL_FRAMEBUFFER, current_framebuffer->framebuffer);
    } else {
        glBindFramebuffer(GL_FRAMEBUFFER, opengl_get_default_framebuffer());
    }
    if (opengl_current_texture_target == GL_TEXTURE_2D) {
        glBindTexture(GL_TEXTURE_2D, opengl_current_texture_id);
    } else {
        glBindTexture(GL_TEXTURE_2D, 0);
    }
    opengl_apply_viewport();
    opengl_apply_clip_region();
    return result;
}

/*-----------------------------------------------------------------------*/

static void update_texture(SysTexture *texture, uint8_t *pixels)
{
    PRECOND(pixels != NULL, return);
    PRECOND(texture != NULL, mem_free(pixels); return);

    glBindTexture(GL_TEXTURE_2D, texture->id);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, texture->width, texture->height, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, pixels);
    if (texture->auto_mipmaps) {
        if (opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)) {
            glGenerateMipmap(GL_TEXTURE_2D);
        } else {
            generate_mipmaps(pixels, 0, texture->width, texture->height, 0);
            pixels = NULL;  // Prevent double free below.
        }
    }
    if (opengl_current_texture_target == GL_TEXTURE_2D) {
        glBindTexture(GL_TEXTURE_2D, opengl_current_texture_id);
    } else {
        glBindTexture(GL_TEXTURE_2D, 0);
    }

    texture->empty = 0;
    texture->color_type = TEXCOLOR_RGBA;
    texture->readable = 1;
    mem_free(pixels);  // In case we didn't call generate_mipmaps().
}

/*-----------------------------------------------------------------------*/

static void generate_mipmaps(uint8_t *pixels, int32_t offset,
                             int width, int height, int level)
{
    PRECOND(pixels != NULL, return);
    PRECOND(width != 0, mem_free(pixels); return);
    PRECOND(height != 0, mem_free(pixels); return);

    if (width == 1 && height == 1) {
        /* No more mipmaps to generate. */
        mem_free(pixels);
        return;
    }

    uint8_t *pixels_base = pixels;
    pixels += offset;

    /* Shrink the texture by half in place, using simple per-component
     * linear averaging.  This algorithm does not handle the case of an
     * odd width or height (other than 1), and mipmaps for such textures
     * will be slightly offset from the primary image. */

    const int new_width  = lbound(width /2, 1);
    const int new_height = lbound(height/2, 1);
    uint8_t *new_pixels = pixels_base;

    if (width == 1 || height == 1) {
        /* Whether horizontal or vertical, the data is still a single
         * array of pixels, and we can treat both cases the same way. */
        for (int i = 0; i < new_width * new_height; i++) {
            for (int c = 0; c < 4; c++) {
                new_pixels[i*4+c] = (pixels[(i*2+0)*4+c]
                                   + pixels[(i*2+1)*4+c] + 1) / 2;
            }
        }
    } else {  // width != 1 && height != 1
        for (int y = 0; y < new_height; y++) {
            const uint8_t *in0 = &pixels[(y*2+0) * (width*4)];
            const uint8_t *in1 = &pixels[(y*2+1) * (width*4)];
            uint8_t *out = &new_pixels[y * (new_width*4)];
            for (int x = 0; x < new_width; x++) {
                for (int c = 0; c < 4; c++) {
                    out[x*4+c] = (in0[(x*2+0)*4+c] + in0[(x*2+1)*4+c]
                                + in1[(x*2+0)*4+c] + in1[(x*2+1)*4+c] + 2) / 4;
                }
            }
        }
    }

    /* Shrink the pixel buffer to reduce memory pressure. */

    const uint32_t new_size = new_width * new_height * 4;
    ASSERT(pixels = mem_realloc(new_pixels, new_size, MEM_ALLOC_TEMP),
           pixels = new_pixels);

    /* Register the mipmap with OpenGL. */

    glTexImage2D(GL_TEXTURE_2D, level+1,
                 GL_RGBA, new_width, new_height, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, pixels);

    /* Recursively generate mipmaps for smaller sizes.  (The initial size
     * check will stop the recursion when we hit 1x1.) */

    generate_mipmaps(pixels, 0, new_width, new_height, level+1);
}

/*************************************************************************/
/**************** Local routines: Pixel format conversion ****************/
/*************************************************************************/

static void convert_bgra8888_rgba8888(void *dest_, const void *src_, size_t len)
{
    const uint32_t *src = (const uint32_t *)src_;
    uint32_t *dest = (uint32_t *)dest_;
    for (size_t i = 0; i < len; i++) {
        const uint32_t pixel = src[i];
#ifdef IS_LITTLE_ENDIAN
        dest[i] = (pixel & 0xFF00FF00)
                | (pixel>>16 & 0xFF)
                | (pixel & 0xFF) << 16;
#else
        dest[i] = (pixel & 0x00FF00FF)
                | (pixel>>16 & 0xFF00)
                | (pixel & 0xFF00) << 16;
#endif
    }
}

/*-----------------------------------------------------------------------*/

static void convert_565rev_565(void *dest_, const void *src_, size_t len)
{
    const uint16_t *src = (const uint16_t *)src_;
    uint16_t *dest = (uint16_t *)dest_;
    for (size_t i = 0; i < len; i++) {
        const uint16_t pixel = src[i];
        dest[i] = pixel>>11 | (pixel & 0x07E0) | pixel<<11;
    }
}

/*-----------------------------------------------------------------------*/

static void convert_rgba1555rev_rgba5551(void *dest_, const void *src_,
                                         size_t len)
{
    const uint16_t *src = (const uint16_t *)src_;
    uint16_t *dest = (uint16_t *)dest_;
    for (size_t i = 0; i < len; i++) {
        const uint16_t pixel = src[i];
        dest[i] = pixel >> 15
                | (pixel & 0x7C00) >> 9
                | (pixel & 0x03E0) << 1
                | pixel << 11;
    }
}

/*-----------------------------------------------------------------------*/

static void convert_bgra1555rev_rgba5551(void *dest_, const void *src_,
                                         size_t len)
{
    const uint16_t *src = (const uint16_t *)src_;
    uint16_t *dest = (uint16_t *)dest_;
    for (size_t i = 0; i < len; i++) {
        const uint16_t pixel = src[i];
        dest[i] = pixel>>15 | pixel<<1;
    }
}

/*-----------------------------------------------------------------------*/

static void convert_rgba4444rev_rgba4444(void *dest_, const void *src_,
                                         size_t len)
{
    const uint16_t *src = (const uint16_t *)src_;
    uint16_t *dest = (uint16_t *)dest_;
    for (size_t i = 0; i < len; i++) {
        const uint16_t pixel = src[i];
        dest[i] = pixel >> 12
                | (pixel & 0x0F00) >> 4
                | (pixel & 0x00F0) << 4
                | pixel << 12;
    }
}

/*-----------------------------------------------------------------------*/

static void convert_bgra4444rev_rgba4444(void *dest_, const void *src_,
                                         size_t len)
{
    const uint16_t *src = (const uint16_t *)src_;
    uint16_t *dest = (uint16_t *)dest_;
    for (size_t i = 0; i < len; i++) {
        const uint16_t pixel = src[i];
        dest[i] = pixel>>12 | pixel<<4;
    }
}

/*-----------------------------------------------------------------------*/

static void convert_a8_rgba8888(void *dest_, const void *src_, size_t len)
{
    const uint8_t *src = (const uint8_t *)src_;
    uint32_t *dest = (uint32_t *)dest_;
    for (size_t i = 0; i < len; i++) {
        const uint32_t pixel = src[i];
#ifdef IS_LITTLE_ENDIAN
        dest[i] = 0x00FFFFFF | pixel<<24;
#else
        dest[i] = 0xFFFFFF00 | pixel;
#endif
    }
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
