/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/psp/thread.c: Thread management routines for the PSP.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/sysdep.h"
#include "src/sysdep/psp/internal.h"
#include "src/sysdep/psp/thread.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Parameter structure for thread_runner(). */
typedef struct ThreadRunParam ThreadRunParam;
struct ThreadRunParam {
    int (*func)(void *arg);  // Function to call.
    void *arg;               // Argument to pass to func().
};

/*-----------------------------------------------------------------------*/

/* Lock nesting count for psp_threads_lock(). */
static unsigned int lock_count;

/* Saved interrupt register value for thread locking. */
static unsigned int lock_intstatus;

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * thread_runner:  Wrapper for threads started with sys_thread_create()
 * which handles interface translation between sysdep and the PSP system
 * call.
 *
 * [Parameters]
 *     args: Thread argument size (== sizeof(ThreadRunParam)).
 *     argp: Thread argument pointer (points to a ThreadRunParam structure).
 * [Return value]
 *     Thread function return value.
 */
static int thread_runner(SceSize args, void *argp);

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

SysThreadID sys_thread_create(
    int priority, int stack_size, int (*function)(void *), void *param)
{
    static unsigned int counter;
    char namebuf[28];
    strformat(namebuf, sizeof(namebuf), "SysThread_%u", counter++);

    int real_priority = sceKernelGetThreadCurrentPriority();
    if (priority != 0) {
        real_priority -= priority; //On the PSP, lower value = higher priority.
        if (real_priority < THREADPRI_USER_MIN) {
            DLOG("Thread priority %d out of range, forcing to %d",
                 real_priority, THREADPRI_USER_MIN);
            real_priority = THREADPRI_USER_MIN;
        } else if (real_priority > THREADPRI_USER_MAX) {
            DLOG("Thread priority %d out of range, forcing to %d",
                 real_priority, THREADPRI_USER_MAX);
            real_priority = THREADPRI_USER_MAX;
        }
    }

    ThreadRunParam trp = {.func = function, .arg = param};
    SceUID thread = psp_start_thread(
        namebuf, thread_runner, real_priority,
        stack_size ? stack_size : 65536, sizeof(trp), &trp);
    if (thread < 0) {
        DLOG("Failed to start thread: %s", psp_strerror(thread));
        return 0;
    } else {
        return thread;
    }
}

/*-----------------------------------------------------------------------*/

int sys_thread_is_running(SysThreadID thread)
{
    SceKernelThreadInfo thinfo;
    mem_clear(&thinfo, sizeof(thinfo));
    thinfo.size = sizeof(thinfo);
    int res = sceKernelReferThreadStatus(thread, &thinfo);

    if (UNLIKELY(res < 0)) {
        DLOG("sceKernelReferThreadStatus(0x%08X) failed: %s",
             thread, psp_strerror(res));
        /* Return false (stopped) so the caller doesn't get stuck. */
        return 0;
    } else {
        return (thinfo.status & (PSP_THREAD_RUNNING | PSP_THREAD_READY
                                 | PSP_THREAD_WAITING)) != 0;
    }
}

/*-----------------------------------------------------------------------*/

int sys_thread_wait(SysThreadID thread, int *result_ret)
{
    if (thread == (SysThreadID)sceKernelGetThreadId()) {
        DLOG("Attempted to wait for current thread!");
        return 0;
    }

    while (!psp_delete_thread_if_stopped(thread, result_ret)) {
        sceKernelDelayThread(1000);  // 1ms
    }
    return 1;
}

/*-----------------------------------------------------------------------*/

void sys_thread_yield()
{
    /* The PSP doesn't have a "yield" function, but this has the same
     * effect.  We can't pass 0 because if we do, the OS will sometimes
     * decide not to switch to a different thread even if one is ready. */
    sceKernelDelayThread(1);
}

/*************************************************************************/
/******************** PSP-specific interface routines ********************/
/*************************************************************************/

SceUID psp_start_thread(const char *name, int (*entry)(SceSize, void *),
                        int priority, int stacksize, SceSize args, void *argp)
{
    if (UNLIKELY(!name)
     || UNLIKELY(!entry)
     || UNLIKELY(priority < 0)
     || UNLIKELY(stacksize < 0)) {
        DLOG("Invalid parameters: %p[%s] %p %d %d %d %p",
             name, name ? name : "", entry, priority, stacksize, args, argp);
        return PSP_EINVAL;
    }

    SceUID handle = sceKernelCreateThread(name, entry, priority, stacksize,
                                          0, NULL);
    if (UNLIKELY(handle < 0)) {
        DLOG("Failed to create thread \"%s\": %s", name, psp_strerror(handle));
        return handle;
    }

    int32_t res = sceKernelStartThread(handle, args, argp);
    if (UNLIKELY(res < 0)) {
        DLOG("Failed to start thread \"%s\": %s", name, psp_strerror(res));
        sceKernelDeleteThread(handle);
        return res;
    }

    return handle;
}

/*-----------------------------------------------------------------------*/

int psp_delete_thread_if_stopped(SceUID thid, int *status_ret)
{
    SceKernelThreadInfo thinfo;
    mem_clear(&thinfo, sizeof(thinfo));
    thinfo.size = sizeof(thinfo);
    int res = sceKernelReferThreadStatus(thid, &thinfo);

    if (UNLIKELY(res < 0)) {
        DLOG("sceKernelReferThreadStatus(0x%08X) failed: %s",
             thid, psp_strerror(res));
        sceKernelTerminateThread(thid);

    } else if (thinfo.status & (PSP_THREAD_RUNNING | PSP_THREAD_READY | PSP_THREAD_WAITING)) {
        return 0;

    } else if (thinfo.status & PSP_THREAD_STOPPED) {
        res = thinfo.exitStatus;

    } else {
        res = 0x80000000 | thinfo.status;
        sceKernelTerminateThread(thid);
    }

    sceKernelDeleteThread(thid);
    if (status_ret) {
        *status_ret = res;
    }
    return 1;
}

/*-----------------------------------------------------------------------*/

void psp_threads_lock(void)
{
    const int intstatus = sceKernelCpuSuspendIntr();
    /* BARRIER() is technically unnecessary on the PSP as long as there's
     * a sequence point (since the PSP's CPU has only one core), but we
     * include one anyway for clarity. */
    BARRIER();
    if (lock_count == 0) {
        lock_intstatus = intstatus;
    }
    lock_count++;
}

/*-----------------------------------------------------------------------*/

void psp_threads_unlock(void)
{
    if (lock_count > 0) {
        lock_count--;
        if (lock_count == 0) {
            BARRIER();  // As above, unnecessary but included for clarity.
            sceKernelCpuResumeIntrWithSync(lock_intstatus);
        }
    }
}

/*-----------------------------------------------------------------------*/

int psp_threads_locked(void)
{
    return lock_count > 0;
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static int thread_runner(SceSize args, void *argp)
{
    if (UNLIKELY(args != sizeof(ThreadRunParam))) {
        DLOG("Invalid argument size %u (should be %u)", args,
             sizeof(ThreadRunParam));
        return -1;
    }
    ThreadRunParam *trp = (ThreadRunParam *)argp;
    return (*trp->func)(trp->arg);
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
